@@ -1,10 +1,12 @@
 package main
 
 import (
+	"errors"
 	"fmt"
 	"math/rand"
 	"net"
 	"strings"
+	"sync"
 
 	"github.com/jedisct1/dlog"
 	"github.com/lifenjoiner/dhcpdns"
@@ -33,6 +35,12 @@ type PluginForward struct {
 	forwardMap         []PluginForwardEntry
 	bootstrapResolvers []string
 	dhcpdns            []*dhcpdns.Detector
+
+	// Hot-reloading support
+	rwLock        sync.RWMutex
+	configFile    string
+	configWatcher *ConfigWatcher
+	stagingMap    []PluginForwardEntry
 }
 
 func (plugin *PluginForward) Name() string {
@@ -44,16 +52,50 @@ func (plugin *PluginForward) Description() string {
 }
 
 func (plugin *PluginForward) Init(proxy *Proxy) error {
-	dlog.Noticef("Loading the set of forwarding rules from [%s]", proxy.forwardFile)
+	plugin.configFile = proxy.forwardFile
+	dlog.Noticef("Loading the set of forwarding rules from [%s]", plugin.configFile)
 
 	if proxy.xTransport != nil {
 		plugin.bootstrapResolvers = proxy.xTransport.bootstrapResolvers
 	}
 
-	lines, err := ReadTextFile(proxy.forwardFile)
+	lines, err := ReadTextFile(plugin.configFile)
 	if err != nil {
 		return err
 	}
+
+	requiresDHCP, forwardMap, err := plugin.parseForwardFile(lines)
+	if err != nil {
+		return err
+	}
+
+	plugin.forwardMap = forwardMap
+
+	if requiresDHCP {
+		if len(proxy.userName) > 0 {
+			dlog.Warn("DHCP/DNS detection may not work when 'user_name' is set or when starting as a non-root user")
+		}
+		if proxy.SourceIPv6 {
+			dlog.Notice("Starting a DHCP/DNS detector for IPv6")
+			d6 := &dhcpdns.Detector{RemoteIPPort: "[2001:DB8::53]:80"}
+			go d6.Serve(9, 10)
+			plugin.dhcpdns = append(plugin.dhcpdns, d6)
+		}
+		if proxy.SourceIPv4 {
+			dlog.Notice("Starting a DHCP/DNS detector for IPv4")
+			d4 := &dhcpdns.Detector{RemoteIPPort: "192.0.2.53:80"}
+			go d4.Serve(9, 10)
+			plugin.dhcpdns = append(plugin.dhcpdns, d4)
+		}
+	}
+	return nil
+}
+
+// parseForwardFile parses forward rules from text
+func (plugin *PluginForward) parseForwardFile(lines string) (bool, []PluginForwardEntry, error) {
+	requiresDHCP := false
+	forwardMap := []PluginForwardEntry{}
+
 	for lineNo, line := range strings.Split(lines, "\n") {
 		line = TrimAndStripInlineComments(line)
 		if len(line) == 0 {
@@ -65,20 +107,19 @@ func (plugin *PluginForward) Init(proxy *Proxy) error {
 			ok = false
 		}
 		if !ok {
-			return fmt.Errorf(
+			return false, nil, fmt.Errorf(
 				"Syntax error for a forwarding rule at line %d. Expected syntax: example.com 9.9.9.9,8.8.8.8",
 				1+lineNo,
 			)
 		}
 		domain = strings.ToLower(domain)
-		requiresDHCP := false
 		var sequence []SearchSequenceItem
 		for _, server := range strings.Split(serversStr, ",") {
 			server = strings.TrimSpace(server)
 			switch server {
 			case "$BOOTSTRAP":
 				if len(plugin.bootstrapResolvers) == 0 {
-					return fmt.Errorf(
+					return false, nil, fmt.Errorf(
 						"Syntax error for a forwarding rule at line %d. No bootstrap resolvers available",
 						1+lineNo,
 					)
@@ -102,68 +143,112 @@ func (plugin *PluginForward) Init(proxy *Proxy) error {
 					dlog.Criticalf("Unknown keyword [%s] at line %d", server, 1+lineNo)
 					continue
 				}
-				if server, err = normalizeIPAndOptionalPort(server, "53"); err != nil {
+				if server, err := normalizeIPAndOptionalPort(server, "53"); err != nil {
 					dlog.Criticalf("Syntax error for a forwarding rule at line %d: %s", 1+lineNo, err)
 					continue
-				}
-				idxServers := -1
-				for i, item := range sequence {
-					if item.typ == Explicit {
-						idxServers = i
-					}
-				}
-				if idxServers == -1 {
-					sequence = append(sequence, SearchSequenceItem{typ: Explicit, servers: []string{server}})
 				} else {
-					sequence[idxServers].servers = append(sequence[idxServers].servers, server)
+					idxServers := -1
+					for i, item := range sequence {
+						if item.typ == Explicit {
+							idxServers = i
+						}
+					}
+					if idxServers == -1 {
+						sequence = append(sequence, SearchSequenceItem{typ: Explicit, servers: []string{server}})
+					} else {
+						sequence[idxServers].servers = append(sequence[idxServers].servers, server)
+					}
+					dlog.Infof("Forwarding [%s] to [%s]", domain, server)
 				}
-				dlog.Infof("Forwarding [%s] to [%s]", domain, server)
 			}
 		}
-		if requiresDHCP {
-			if len(proxy.userName) > 0 {
-				dlog.Warn("DHCP/DNS detection may not work when 'user_name' is set or when starting as a non-root user")
-			}
-			if proxy.SourceIPv6 {
-				dlog.Notice("Starting a DHCP/DNS detector for IPv6")
-				d6 := &dhcpdns.Detector{RemoteIPPort: "[2001:DB8::53]:80"}
-				if err := d6.Detect(); err != nil {
-					dlog.Criticalf("Failed to start the DHCP/DNS IPv6 server: %s", err)
-					continue
-				}
-				go d6.Serve(9, 10)
-				plugin.dhcpdns = append(plugin.dhcpdns, d6)
-			}
-			if proxy.SourceIPv4 {
-				dlog.Notice("Starting a DHCP/DNS detector for IPv4")
-				d4 := &dhcpdns.Detector{RemoteIPPort: "192.0.2.53:80"}
-				if err := d4.Detect(); err != nil {
-					dlog.Criticalf("Failed to start the DHCP/DNS IPv4 server: %s", err)
-					continue
-				}
-				go d4.Serve(9, 10)
-				plugin.dhcpdns = append(plugin.dhcpdns, d4)
-			}
-		}
-		plugin.forwardMap = append(plugin.forwardMap, PluginForwardEntry{
+		forwardMap = append(forwardMap, PluginForwardEntry{
 			domain:   domain,
 			sequence: sequence,
 		})
 	}
-	return nil
+
+	return requiresDHCP, forwardMap, nil
 }
 
 func (plugin *PluginForward) Drop() error {
+	if plugin.configWatcher != nil {
+		plugin.configWatcher.RemoveFile(plugin.configFile)
+	}
 	return nil
 }
 
-func (plugin *PluginForward) Reload() error {
+// PrepareReload loads new rules into staging structure but doesn't apply them yet
+func (plugin *PluginForward) PrepareReload() error {
+	// Read the configuration file
+	lines, err := SafeReadTextFile(plugin.configFile)
+	if err != nil {
+		return fmt.Errorf("error reading config file during reload preparation: %w", err)
+	}
+
+	// Parse the forward file
+	_, stagingMap, err := plugin.parseForwardFile(lines)
+	if err != nil {
+		return fmt.Errorf("error parsing config during reload preparation: %w", err)
+	}
+
+	// Store in staging area
+	plugin.stagingMap = stagingMap
+
 	return nil
 }
 
+// ApplyReload atomically replaces the active rules with the staging ones
+func (plugin *PluginForward) ApplyReload() error {
+	if plugin.stagingMap == nil {
+		return errors.New("no staged configuration to apply")
+	}
+
+	// Use write lock to swap rule structures
+	plugin.rwLock.Lock()
+	plugin.forwardMap = plugin.stagingMap
+	plugin.stagingMap = nil
+	plugin.rwLock.Unlock()
+
+	dlog.Noticef("Applied new configuration for plugin [%s]", plugin.Name())
+	return nil
+}
+
+// CancelReload cleans up any staging resources
+func (plugin *PluginForward) CancelReload() {
+	plugin.stagingMap = nil
+}
+
+// Reload implements hot-reloading for the plugin
+func (plugin *PluginForward) Reload() error {
+	dlog.Noticef("Reloading configuration for plugin [%s]", plugin.Name())
+
+	// Prepare the new configuration
+	if err := plugin.PrepareReload(); err != nil {
+		plugin.CancelReload()
+		return err
+	}
+
+	// Apply the new configuration
+	return plugin.ApplyReload()
+}
+
+// GetConfigPath returns the path to the plugin's configuration file
+func (plugin *PluginForward) GetConfigPath() string {
+	return plugin.configFile
+}
+
+// SetConfigWatcher sets the config watcher for this plugin
+func (plugin *PluginForward) SetConfigWatcher(watcher *ConfigWatcher) {
+	plugin.configWatcher = watcher
+}
+
 func (plugin *PluginForward) Eval(pluginsState *PluginsState, msg *dns.Msg) error {
 	qName := pluginsState.qName
 	qNameLen := len(qName)
+
+	// Use read lock for thread-safe access to forwardMap
+	plugin.rwLock.RLock()
 	var sequence []SearchSequenceItem
 	for _, candidate := range plugin.forwardMap {
 		candidateLen := len(candidate.domain)
@@ -177,6 +262,8 @@ func (plugin *PluginForward) Eval(pluginsState *PluginsState, msg *dns.Msg) erro
 			break
 		}
 	}
+	plugin.rwLock.RUnlock()
+
 	if len(sequence) == 0 {
 		return nil
 	}
@@ -219,23 +306,22 @@ func (plugin *PluginForward) Eval(pluginsState *PluginsState, msg *dns.Msg) erro
 		tries--
 		dlog.Debugf("Forwarding [%s] to [%s]", qName, server)
 		client := dns.Client{Net: pluginsState.serverProto, Timeout: pluginsState.timeout}
-		respMsg, _, err = client.Exchange(msg, server)
+
+		// Create a clean copy of the message without Extra section for forwarding
+		forwardMsg := msg.Copy()
+		forwardMsg.Extra = nil
+
+		respMsg, _, err = client.Exchange(forwardMsg, server)
 		if err != nil {
 			continue
 		}
 		if respMsg.Truncated {
 			client.Net = "tcp"
-			respMsg, _, err = client.Exchange(msg, server)
+			respMsg, _, err = client.Exchange(forwardMsg, server)
 			if err != nil {
 				continue
 			}
 		}
-		if len(sequence) > 0 {
-			switch respMsg.Rcode {
-			case dns.RcodeNameError, dns.RcodeRefused, dns.RcodeNotAuth:
-				continue
-			}
-		}
 		if edns0 := respMsg.IsEdns0(); edns0 == nil || !edns0.Do() {
 			respMsg.AuthenticatedData = false
 		}
@@ -243,6 +329,12 @@ func (plugin *PluginForward) Eval(pluginsState *PluginsState, msg *dns.Msg) erro
 		pluginsState.synthResponse = respMsg
 		pluginsState.action = PluginsActionSynth
 		pluginsState.returnCode = PluginsReturnCodeForward
+		if len(sequence) > 0 {
+			switch respMsg.Rcode {
+			case dns.RcodeNameError, dns.RcodeRefused, dns.RcodeNotAuth:
+				continue
+			}
+		}
 		return nil
 	}
 	return err
